from unittest import TestCase
import warnings

from paper_cup.client import SNSClient, SQSClient
from paper_cup.paper_cup import PaperCup


class TestPaperCup(TestCase):

  def setUp(self):
    """Silent the warnings as explained here:
        https://github.com/boto/boto3/issues/454
      with solution from here:
        https://stackoverflow.com/questions/26563711/disabling-python-3-2-resourcewarning
    """
    warnings.simplefilter("ignore", ResourceWarning)

  def tearDown(self):
    """Put back the warnings."""
    warnings.simplefilter("default", ResourceWarning)

  def test_instance(self):
    """Check that the instance initialize correctly."""
    # We need to have a topic to connect to
    SNSClient(PaperCup.PC_AWS_LOCAL_ENDPOINT).create_topic(PaperCup.PC_TOPIC)
    paper_cup = PaperCup()

    # check that we correctly set sns and sqs sessions
    self.assertTrue(paper_cup.sns)
    self.assertTrue(paper_cup.topic_arn)
    self.assertTrue(paper_cup.sqs)

  def test_bulk_publish(self):
    """Check that it all the messages are sent."""
    import json
    # list of one message
    list_message = [DummyMessage()]
    list_action = ['index']
    paper_cup = pc_instance()
    paper_cup.bulk_publish(list_message, list_action)

    # FIXME remove this after refactor the constructors
    paper_cup.sqs_queue = paper_cup.sqs.get_queue_by_name(paper_cup.PC_QUEUE)

    # Get the Message
    sqs_msgs = paper_cup.sqs_queue.receive_messages()
    sqs_msgs_dict = json.loads(sqs_msgs[0].body)

    msg = json.loads(sqs_msgs_dict['Message'])
    # check that Message content is a list of messages
    self.assertTrue(isinstance(msg, list))

    # clean the queue and topic
    clean_after(paper_cup)

  def test_bulk_consume(self):
    """"Check that we can read bulk data."""

    list_message = [DummyMessage(), DummyMessage(qwe="qwe", abc="abc", number=1)]
    list_action = ['index', 'delete']
    paper_cup = pc_instance()
    paper_cup.bulk_publish(list_message, list_action)

    list_result = []

    class ConsumePaperCup(PaperCup):
      """Dummy class to have consume in the name"""

      def index(self, message):
        """Dummy method to test the consume action."""
        list_result.append(message)

      def delete(self, message):
        """Dummy method to test the consume action."""
        list_result.append(message)

    PaperCup().run()
    # check that we have the 2 messages in the list i.e that we consume correclty the two actions
    self.assertEqual(len(list_result), 2)


# ################ Other utils for test maybe better to move to their own file


def pc_instance():
  """Build and return a ready to use paper cup instance."""
  SNSClient(PaperCup.PC_AWS_LOCAL_ENDPOINT).create_topic(PaperCup.PC_TOPIC)

  class PublishPaperCup(PaperCup):
    """Dummy class to have Publish key in the name."""

  pc_instance = PublishPaperCup()

  # delete the queue if the unit test fail the queue may be here
  try:
    pc_instance.sqs.delete_queue(PaperCup.PC_QUEUE)
  except Exception:
    # if the queue don't exist it will raise but we don't care
    pass

  SQSClient(PaperCup.PC_AWS_LOCAL_ENDPOINT).create_queue(PaperCup.PC_QUEUE)

  pc_instance.topic_arn = pc_instance.sns.get_topic_arn(PaperCup.PC_TOPIC)
  pc_instance.queue_url = pc_instance.sqs.get_queue_url(PaperCup.PC_QUEUE)
  pc_instance.queue_arn = pc_instance.sqs.get_queue_arn(pc_instance.queue_url)

  # attach the topic to the queue so when we publish the message will go this queue
  pc_instance.sns.subscribe(pc_instance.topic_arn, pc_instance.queue_arn)

  return pc_instance


def clean_after(pc_instance):
  """Delete after usage."""
  pc_instance.sqs.delete_queue(pc_instance.queue_url)
  pc_instance.sns.delete_topic(pc_instance.topic_arn)


class DummyMessage(dict):
  """Build a dummy dict with data inside as the message can be any json valid data."""

  def __init__(self, *args, **kwargs):
    """Set dummy data by default."""
    if not kwargs:
      kwargs = {
          "Band": "Soilwork",
          "Album": "The living Infinity",
          "Tracklist": [
              {"track_num": 1, "title": "Spectrum of Eternity", "duration": "4:00"},
              {"track_num": 2, "title": "Memories Confined", "duration": "3:25"}
          ]
      }
    super().__init__(*args, **kwargs)
    self.__dict__ = self



import boto3


class SNSClient:
  """Common sns usages."""

  def __init__(self, endpoint_url, region='ap-northeast-1', aws_access_key_id=None, aws_secret_access_key=None):
    """Constructor with already set in options."""

    session = boto3.Session(region_name=region, aws_access_key_id=aws_secret_access_key, aws_secret_access_key=aws_secret_access_key)
    self.sns = session.client('sns', endpoint_url=endpoint_url)

  def get_topic_arn(self, topic_name, token=None):
    """Get sns topic by name and paginate until is found."""
    sns = self.sns

    if token:
      list_topics = sns.list_topics(NextToken=token)
    else:
      list_topics = sns.list_topics()

    next_token = list_topics.get('NextToken')
    topic_list = list_topics['Topics']

    for topic in topic_list:
      if topic_name in topic['TopicArn'].split(':')[5]:
        return topic['TopicArn']
    else:
      if next_token:
        self.get_topic(token=next_token)
      else:
        raise NotImplementedError('SNS topic not found!')

  def publish(self, message, topic_arn):
    """Send message to all subscriber of this topic."""
    return self.sns.publish(
        TopicArn=topic_arn,
        Message=message,
    )

  def subscribe(self, topic_arn, queue_arn):
    """Subscribe to topic."""
    return self.sns.subscribe(
        Protocol='sqs',
        TopicArn=topic_arn,
        Endpoint=queue_arn,
    )

  def create_topic(self, topic_name):
    """Create SNS topic by name."""
    return self.sns.create_topic(Name=topic_name)['TopicArn']

  def delete_topic(self, topic_arn):
    """Delete a given SNS topic from it arn."""
    return self.sns.delete_topic(TopicArn=topic_arn)


class SQSClient:
  """Common sqs usages."""

  def __init__(self, endpoint_url, region='ap-northeast-1', aws_access_key_id=None, aws_secret_access_key=None):
    """Constructor with already set in options."""
    session = boto3.Session(region_name=region, aws_access_key_id=aws_secret_access_key, aws_secret_access_key=aws_secret_access_key)
    self.sqs = session.client('sqs', endpoint_url=endpoint_url)
    self.sqs_obj = session.resource('sqs', endpoint_url=endpoint_url)

  def create_queue(self, queue_name):
    """Create SQS queue by name."""
    return self.sqs.create_queue(QueueName=queue_name)

  def delete_queue(self, queue_url):
    """Delete a given SQS queue from it url."""
    return self.sqs.delete_queue(QueueUrl=queue_url)

  def get_queue_by_name(self, queue_name):
    """Get SQS queue object by name."""
    return self.sqs_obj.get_queue_by_name(QueueName=queue_name)

  def get_queue_url(self, queue_name):
    """Get SQS queue url by it name."""
    return self.sqs.get_queue_url(QueueName=queue_name)['QueueUrl']

  def get_queue_arn(self, queue_url):
    """Get SQS queue arn by it url."""
    sqs_queue_attrs = self.sqs.get_queue_attributes(QueueUrl=queue_url, AttributeNames=['All'])['Attributes']
    return sqs_queue_attrs['QueueArn']



import json

from .client import SNSClient, SQSClient


class PaperCup(object):
  """Publisher and subscribe settings."""
  PC_ENABLE = True
  PC_TOPIC = 'topic' # Publisher (SNS)
  PC_QUEUE = 'queue' # Consumer (SQS)
  PC_LISTEN = ['service'] # list of service name for message to process ex: cas listen only 'booking' so for cas PC_LISTEN = ['booking']
  PC_SENDER = 'service' # name of the app that send the message ex: 'booking'

  # default values set for test
  PC_AWS_ACCESS_KEY_ID = 'test'
  PC_AWS_SECRET_ACCESS_KEY_ID = 'test'
  PC_AWS_LOCAL_ENDPOINT = 'http://192.168.56.1:9010' # we use moto

  # set default attribut values
  sns = False
  sqs = False

  def __init__(self):
    """Deprecated As we never need Publish and Consume in same time."""
    if self.PC_ENABLE:
      self.sns = SNSClient(self.PC_AWS_LOCAL_ENDPOINT, aws_access_key_id=self.PC_AWS_ACCESS_KEY_ID, aws_secret_access_key=self.PC_AWS_SECRET_ACCESS_KEY_ID)
      self.topic_arn = self.sns.get_topic_arn(self.PC_TOPIC)

      self.sqs = SQSClient(self.PC_AWS_LOCAL_ENDPOINT, aws_access_key_id=self.PC_AWS_ACCESS_KEY_ID, aws_secret_access_key=self.PC_AWS_SECRET_ACCESS_KEY_ID)

  def publish(self, message, action):
    """Send message to sns."""
    if self.sns:
      message = self._add_more_data(message, action)
      message = json.dumps(message)
      self.sns.publish(message, self.topic_arn)

  def _add_more_data(self, message, action):
    """Add necessary data to detemine the consumer and action function."""
    # we expect the publish class name as PublishUser and it's cosumer will be ConsumeUser
    class_name = self.__class__.__name__
    message['consumer_action_class'] = class_name.replace('Publish', 'Consume')
    message['action'] = action
    message['sender'] = self.PC_SENDER
    return message

  def bulk_publish(self, list_message, list_action):
    """Send message by bulk to sns.
      Please limit the list to 50 messages (Need to be done in the call as it depends on the message).
    """
    message = ''
    msg_list = []
    if self.sns:
      for i, one_message in enumerate(list_message):
        full_message = self._add_more_data(one_message, list_action[i])
        temp_message = json.dumps(full_message)
        # check max size of the message to publish under the limit
        if (len(message) + len(temp_message)) > 256000:
          self.sns.publish(message, self.topic_arn)
          msg_list = [full_message]
        else:
          msg_list.append(one_message)
        message = json.dumps(msg_list)

      if message:
        self.sns.publish(message, self.topic_arn)

  def run(self):
    """Deprecated as bad naming.
      Read the message in queue and use the class that will handle the action. (Consume the message)
    """
    if self.sqs:
      queue = self.sqs.get_queue_by_name(self.PC_QUEUE)
      # get all the consumer classes that will handle actions
      action_classes = {cls.__name__: cls() for cls in self.__class__.__subclasses__() if 'Consume' in cls.__name__}
      messages = queue.receive_messages(WaitTimeSeconds=20, MaxNumberOfMessages=10, VisibilityTimeout=30)
      while messages:
        for message in messages:
          body = json.loads(message.body)
          msg = json.loads(body['Message'])

          if isinstance(msg, list):
            for one_msg in msg:
              self._consume_msg(one_msg, action_classes)
          else:
            self._consume_msg(msg, action_classes)

          message.delete()

        messages = queue.receive_messages(MaxNumberOfMessages=10, VisibilityTimeout=30)

  def _consume_msg(self, msg, action_classes):
    """Common call to consume a message."""
    action = msg.get('action')
    sender = msg.get('sender')

    # check that the consumer listen from the sender and do the action
    if sender in self.PC_LISTEN:
      consumer_action_class = msg.get('consumer_action_class')
      action_class = action_classes.get(consumer_action_class)
      # only handle the action with consumers
      if action_class:
        action_class._do_action(msg, action)

  def _do_action(self, message, action):
    """Call the action of the consumer class."""
    method = getattr(self, action)
    method(message)


class PublishPC(PaperCup):
  """Public class for Publisher."""

  def __init__(self, *args, **kwargs):
    """"""
    if self.PC_ENABLE:
      self.sns = SNSClient(self.PC_AWS_LOCAL_ENDPOINT, aws_access_key_id=self.PC_AWS_ACCESS_KEY_ID, aws_secret_access_key=self.PC_AWS_SECRET_ACCESS_KEY_ID)
      self.topic_arn = self.sns.get_topic_arn(self.PC_TOPIC)


class ConsumePC(PaperCup):
  """Public class for Consume ."""

  def __init__(self, *args, **kwargs):
    """"""
    if self.PC_ENABLE:
      self.sqs = SQSClient(self.PC_AWS_LOCAL_ENDPOINT, aws_access_key_id=self.PC_AWS_ACCESS_KEY_ID, aws_secret_access_key=self.PC_AWS_SECRET_ACCESS_KEY_ID)
      self.sqs_queue = self.sqs.get_queue_by_name(self.PC_QUEUE)

  def consume(self):
    """Read the message in queue and use the class that will handle the action. (Consume the message)"""
    if self.sqs_queue:
      # get all the consumer classes that will handle actions
      action_classes = {cls.__name__: cls() for cls in self.__class__.__subclasses__() if 'Consume' in cls.__name__}
      messages = self.sqs_queue.receive_messages(WaitTimeSeconds=20, MaxNumberOfMessages=10, VisibilityTimeout=30)

      while messages:
        for message in messages:
          body = json.loads(message.body)
          msg = json.loads(body['Message'])

          if isinstance(msg, list):
            for one_msg in msg:
              self._consume_msg(one_msg, action_classes)
          else:
            self._consume_msg(msg, action_classes)

          message.delete()

        messages = self.sqs_queue.receive_messages(MaxNumberOfMessages=10, VisibilityTimeout=30)



from unittest import TestCase

from paper_cup.client import SNSClient, SQSClient
from paper_cup import PaperCup


class TestSNSClient(TestCase):

  def test_instance(self):
    """Check that the instance initialize correctly."""
    sns_client = SNSClient(PaperCup.PC_AWS_LOCAL_ENDPOINT)
    # check that we correctly set an sns session
    self.assertTrue(sns_client.sns)


class TestSQSClient(TestCase):

  def test_instance(self):
    """Check that the instance initialize correctly."""
    sqs_client = SQSClient(PaperCup.PC_AWS_LOCAL_ENDPOINT)
    # check that we correctly set an sqs session
    self.assertTrue(sqs_client.sqs)
    self.assertTrue(sqs_client.sqs_obj)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='paper_cup',
    version='20.8.27',
    license='MIT',
    homepage='https://github.com/LUXEYS/paper_cup',
    author='Luxeys',
    url='https://github.com/LUXEYS/paper_cup',
    author_email='bryan@luxeys.com',
    description='Micro-service communication system powered with paper cup engine!',
    packages=find_packages(),
    install_requires=[
        'boto3',
    ],
    long_description_content_type="text/markdown",
    long_description=long_description,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Topic :: Internet :: WWW/HTTP',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3'
    ],
    zip_safe=False,
)
